#include "i8080.h"

unsigned char ram[65536] = { 0xF2, 0x00, 0xDC, 0x08, 0x00, 0xF2, 0x00, 0xF0 };
unsigned char io[256];
unsigned char regA, regB, regC, regD, regE, regH, regL;
unsigned char regF;
unsigned short regSP, regPC;
unsigned char intr_flag;

void halt(void) {
  regPC = 0;
}

unsigned char fetch(void) {
  return ram[regPC++];
}

void setflags(unsigned short tmp) {
  unsigned short tmp2, p = 0;
  tmp &= 0xFF;
  // S(sign) flag
  if (tmp & 0x80) {
    regF |= flagS;
  } else {
    regF &= ~flagS;
  }
  // Z(zero) flag
  if (tmp == 0) {
    regF |= flagZ;
  } else {
    regF &= ~flagZ;
  }
  // P(parity) flag
  tmp2 = tmp;
  for (int i = 0; i < 8; i++) {
    p += (tmp2 & 1);
    tmp2 >>= 1;
  }
  if (p & 1) {
    regF &= ~flagP;
  } else {
    regF |= flagP;
  }
}

unsigned char add8(unsigned char a, unsigned char b, unsigned char c) {
  unsigned short tmp;
  tmp = a + b + c;
  setflags(tmp);
  // C(Carry) flag
  if (tmp & 0x100) {
    regF |= flagC;
  } else {
    regF &= ~flagC;
  }
  // A(Half Carry) flag
  if (((a & 0xF) + (b & 0xF) + (c & 0xF)) > 0xF) {
    regF |= flagA;
  } else {
    regF &= ~flagA;
  }
  return (unsigned char)(tmp & 0xFF);
}

unsigned char sub8(unsigned char a, unsigned char b, unsigned char c) {
  unsigned short tmp;
  unsigned char m_b, m_c;
  m_b = (unsigned char)(~b);
  m_c = c ? 0 : 1;
  tmp = a + m_b + m_c;
  setflags(tmp);
  // C(Carry) flag
  if (tmp & 0x100) {
    regF &= ~flagC;
  } else {
    regF |= flagC;
  }
  // A(Half Carry) flag
  if (((a & 0xF) + (m_b & 0xF) + m_c) & 0x10) {
    regF |= flagA;
  } else {
    regF &= ~flagA;
  }
  return (unsigned char)(tmp & 0xFF);
}

unsigned char and8(unsigned char a, unsigned char b) {
  unsigned short tmp;
  tmp = a & b;
  setflags(tmp);
  if ((a | b) & 0x08) {
    regF |= flagA;
  } else {
    regF &= ~flagA;
  }
  regF &= ~flagC;
  return tmp;
}

unsigned char xor8(unsigned char a, unsigned char b) {
  unsigned short tmp;
  tmp = a ^ b;
  setflags(tmp);
  regF &= ~flagC;
  regF &= ~flagA;
  return (unsigned char)(tmp & 0xFF);
}

unsigned char or8(unsigned char a, unsigned char b) {
  unsigned short tmp;
  tmp = a | b;
  setflags(tmp);
  regF &= ~flagC;
  regF &= ~flagA;
  return tmp;
}

unsigned short add16(unsigned short a, unsigned short b) {
  unsigned int tmp, ac;
  tmp = a + b;
  // C(carry) flag
  if (tmp & 0x10000) {
    regF |= flagC;
  } else {
    regF &= ~flagC;
  }
  return (unsigned short)(tmp & 0xFFFF);
}

void do_nop(unsigned char opcode) {
}

void do_lxi(unsigned char opcode) {
  unsigned char high, low;
  switch (opcode) {
    case 0x01:
      regC = fetch();
      regB = fetch();
      break;
    case 0x11:
      regE = fetch();
      regD = fetch();
      break;
    case 0x21:
      regL = fetch();
      regH = fetch();
      break;
    case 0x31:
      low = fetch();
      high = fetch();
      regSP = (high << 8) + low;
      break;
    default:
      break;
  }
}

void do_sta(unsigned char opcode) {
  unsigned char high, low;
  unsigned short addr;
  switch (opcode) {
    case 0x02:
      ram[regBC] = regA;
      break;
    case 0x12:
      ram[regDE] = regA;
      break;
    case 0x22:
      low = fetch();
      high = fetch();
      addr = (high << 8) + low;
      ram[addr] = regL;
      ram[addr + 1] = regH;
      break;
    case 0x32:
      low = fetch();
      high = fetch();
      ram[(high << 8) + low] = regA;
      break;
    default:
      break;
  }
}

void do_inx(unsigned char opcode) {
  unsigned short tmp = 0;
  switch (opcode) {
    case 0x03:
      tmp = regBC + 1;
      regC = tmp & 0xFF;
      regB = (tmp >> 8) & 0xFF;
      break;
    case 0x13:
      tmp = regDE + 1;
      regE = tmp & 0xFF;
      regD = (tmp >> 8) & 0xFF;
      break;
    case 0x23:
      tmp = regHL + 1;
      regL = tmp & 0xFF;
      regH = (tmp >> 8) & 0xFF;
      break;
    case 0x33:
      regSP++;
      break;
    default:
      break;
  }
}

void do_dcx(unsigned char opcode) {
  unsigned short tmp = 0;
  switch (opcode) {
    case 0x0B:
      tmp = regBC - 1;
      regC = tmp & 0xFF;
      regB = (tmp >> 8) & 0xFF;
      break;
    case 0x1B:
      tmp = regDE - 1;
      regE = tmp & 0xFF;
      regD = (tmp >> 8) & 0xFF;
      break;
    case 0x2B:
      tmp = regHL - 1;
      regL = tmp & 0xFF;
      regH = (tmp >> 8) & 0xFF;
      break;
    case 0x3B:
      regSP--;
      break;
    default:
      break;
  }
}

void do_inr(unsigned char opcode) {
  unsigned char f_bak;
  unsigned short addr;

  f_bak = regF;   // reserve flags
  switch (opcode) {
    case 0x04:
      regB = add8(regB, 1, 0);
      break;
    case 0x0C:
      regC = add8(regC, 1, 0);
      break;
    case 0x14:
      regD = add8(regD, 1, 0);
      break;
    case 0x1C:
      regE = add8(regE, 1, 0);
      break;
    case 0x24:
      regH = add8(regH, 1, 0);
      break;
    case 0x2C:
      regL = add8(regL, 1, 0);
      break;
    case 0x34:
      ram[regHL] = add8(ram[regHL], 1, 0);
      break;
    case 0x3C:
      regA = add8(regA, 1, 0);
      break;
    default:
      break;
  }
  regF &= ~flagC;
  regF |= (f_bak & flagC);
}

void do_dcr(unsigned char opcode) {
  unsigned char f_bak;

  f_bak = regF;   // reserve flags
  switch (opcode) {
    case 0x05:
      regB = sub8(regB, 1, 0);
      break;
    case 0x0D:
      regC = sub8(regC, 1, 0);
      break;
    case 0x15:
      regD = sub8(regD, 1, 0);
      break;
    case 0x1D:
      regE = sub8(regE, 1, 0);
      break;
    case 0x25:
      regH = sub8(regH, 1, 0);
      break;
    case 0x2D:
      regL = sub8(regL, 1, 0);
      break;
    case 0x35:
      ram[regHL] = sub8(ram[regHL], 1, 0);
      break;
    case 0x3D:
      regA = sub8(regA, 1, 0);
      break;
    default:
      break;
  }
  regF &= ~flagC;
  regF |= (f_bak & flagC);
}

void do_mvi(unsigned char opcode) {
  unsigned char data;

  data = fetch();
  switch (opcode) {
    case 0x06:
      regB = data;
      break;
    case 0x0E:
      regC = data;
      break;
    case 0x16:
      regD = data;
      break;
    case 0x1E:
      regE = data;
      break;
    case 0x26:
      regH = data;
      break;
    case 0x2E:
      regL = data;
      break;
    case 0x36:
      ram[regHL] = data;
      break;
    case 0x3E:
      regA = data;
      break;
    default:
      break;
  }
}

void do_rlc(unsigned char opcode) {
  unsigned char f_bak, add;
  switch (opcode) {
    case 0x07:
      if (regA & 0x80) {
	regA <<= 1;
	regA |= 1;
	regF |= flagC;
      } else {
	regA <<= 1;
	regF &= ~flagC;
      }
      break;
    case 0x0F:
      if (regA & 0x01) {
	regA >>= 1;
	regA |= 0x80;
	regF |= flagC;
      } else {
	regA >>= 1;
	regF &= ~flagC;
      }
      break;
    case 0x17:
      f_bak = regF;
      if (regA & 0x80) {
	regA <<= 1;
	regF |= flagC;
      } else {
	regA <<= 1;
	regF &= ~flagC;
      }
      regA |= (f_bak & flagC);
      break;
    case 0x1F:
      f_bak = regF;
      if (regA & 0x01) {
	regA >>= 1;
	regF |= flagC;
      } else {
	regA >>= 1;
	regF &= ~flagC;
      }
      regA |= ((f_bak & flagC) << 7);
      break;
    case 0x27:     //DAA
      f_bak = regF;
      add = 0;
      if ((regF & flagA) || ((regA & 0x0F) > 9)) {
	add |= 0x06;
      }
      if ((regF & flagC) ||
	  ((regA & 0xF0) >= 0x90) && ((regA & 0x0F) > 9) ||
	  ((regA & 0xF0) > 0x90)) {
	add |= 0x60;
	f_bak |= flagC;
      }
      regA = add8(regA, add, 0);
      regF &= ~flagC;
      regF |= (f_bak & flagC);
      break;
    case 0x2F:
      regA = (unsigned char)(~regA & 0xFF);
      break;
    case 0x37:
      regF |= flagC;
      break;
    case 0x3F:
      if (regF & flagC) {
	regF &= ~flagC;
      } else {
	regF |= flagC;
      }
      break;
    default:
      break;
  }
}

void do_dad(unsigned char opcode) {
  unsigned short tmp = 0;

  switch (opcode) {
    case 0x09:
      tmp  = add16(regHL, regBC);
      break;
    case 0x19:
      tmp  = add16(regHL, regDE);
      break;
    case 0x29:
      tmp  = add16(regHL, regHL);
      break;
    case 0x39:
      tmp  = add16(regHL, regSP);
      break;
    default:
      break;
  }
  regL = (unsigned char)(tmp & 0xFF);
  regH = (unsigned char)(tmp >> 8);
}

void do_lda(unsigned char opcode) {
  unsigned char high, low;
  unsigned short addr;
  switch (opcode) {
    case 0x0A:
      regA = ram[regBC];
      break;
    case 0x1A:
      regA = ram[regDE];
      break;
    case 0x2A:
      low = fetch();
      high = fetch();
      addr = (high << 8) + low;
      regL = ram[addr];
      regH = ram[addr + 1];
      break;
    case 0x3A:
      low = fetch();
      high = fetch();
      addr = (high << 8) + low;
      regA = ram[addr];
    default:
      break;
  }
}

void do_movb(unsigned char opcode) {
  switch (opcode) {
    case 0x40:
      break;
    case 0x41:
      regB = regC;
      break;
    case 0x42:
      regB = regD;
      break;
    case 0x43:
      regB = regE;
      break;
    case 0x44:
      regB = regH;
      break;
    case 0x45:
      regB = regL;
      break;
    case 0x46:
      regB = ram[regHL];
      break;
    case 0x47:
      regB = regA;
      break;
    default:
      break;
  }
}

void do_movc(unsigned char opcode) {
  switch (opcode) {
    case 0x48:
      regC = regB;
      break;
    case 0x49:
      break;
    case 0x4A:
      regC = regD;
      break;
    case 0x4B:
      regC = regE;
      break;
    case 0x4C:
      regC = regH;
      break;
    case 0x4D:
      regC = regL;
      break;
    case 0x4E:
      regC = ram[regHL];
      break;
    case 0x4F:
      regC = regA;
      break;
    default:
      break;
  }
}

void do_movd(unsigned char opcode) {
  switch (opcode) {
    case 0x50:
      regD = regB;
      break;
    case 0x51:
      regD = regC;
      break;
    case 0x52:
      break;
    case 0x53:
      regD = regE;
      break;
    case 0x54:
      regD = regH;
      break;
    case 0x55:
      regD = regL;
      break;
    case 0x56:
      regD = ram[regHL];
      break;
    case 0x57:
      regD = regA;
      break;
    default:
      break;
  }
}

void do_move(unsigned char opcode) {
  switch (opcode) {
    case 0x58:
      regE = regB;
      break;
    case 0x59:
      regE = regC;
      break;
    case 0x5A:
      regE = regD;
      break;
    case 0x5B:
      break;
    case 0x5C:
      regE = regH;
      break;
    case 0x5D:
      regE = regL;
      break;
    case 0x5E:
      regE = ram[regHL];
      break;
    case 0x5F:
      regE = regA;
      break;
    default:
      break;
  }
}

void do_movh(unsigned char opcode) {
  switch (opcode) {
    case 0x60:
      regH = regB;
      break;
    case 0x61:
      regH = regC;
      break;
    case 0x62:
      regH = regD;
      break;
    case 0x63:
      regH = regE;
      break;
    case 0x64:
      break;
    case 0x65:
      regH = regL;
      break;
    case 0x66:
      regH = ram[regHL];
      break;
    case 0x67:
      regH = regA;
      break;
    default:
      break;
  }
}

void do_movl(unsigned char opcode) {
  switch (opcode) {
    case 0x68:
      regL = regB;
      break;
    case 0x69:
      regL = regC;
      break;
    case 0x6A:
      regL = regD;
      break;
    case 0x6B:
      regL = regE;
      break;
    case 0x6C:
      regL = regH;
      break;
    case 0x6D:
      break;
    case 0x6E:
      regL = ram[regHL];
      break;
    case 0x6F:
      regL = regA;
      break;
    default:
      break;
  }
}

void do_movm(unsigned char opcode) {
  switch (opcode) {
    case 0x70:
      ram[regHL] = regB;
      break;
    case 0x71:
      ram[regHL] = regC;
      break;
    case 0x72:
      ram[regHL] = regD;
      break;
    case 0x73:
      ram[regHL] = regE;
      break;
    case 0x74:
      ram[regHL] = regH;
      break;
    case 0x75:
      ram[regHL] = regL;
      break;
    case 0x76:
      halt();
      break;
    case 0x77:
      ram[regHL] = regA;
    default:
      break;
  }
}

void do_mova(unsigned char opcode) {
  switch (opcode) {
    case 0x78:
      regA = regB;
      break;
    case 0x79:
      regA = regC;
      break;
    case 0x7A:
      regA = regD;
      break;
    case 0x7B:
      regA = regE;
      break;
    case 0x7C:
      regA = regH;
      break;
    case 0x7D:
      regA = regL;
      break;
    case 0x7E:
      regA = ram[regHL];
      break;
    case 0x7F:
      break;
    default:
      break;
  }
}

void do_add(unsigned char opcode) {
  unsigned char c;
  
  c = regF & flagC;
  switch (opcode) {
    case 0x80:
      regA = add8(regA, regB, 0);
      break;
    case 0x81:
      regA = add8(regA, regC, 0);
      break;
    case 0x82:
      regA = add8(regA, regD, 0);
      break;
    case 0x83:
      regA = add8(regA, regE, 0);
      break;
    case 0x84:
      regA = add8(regA, regH, 0);
      break;
    case 0x85:
      regA = add8(regA, regL, 0);
      break;
    case 0x86:
      regA = add8(regA, ram[regHL], 0);
      break;
    case 0x87:
      regA = add8(regA, regA, 0);
      break;
    case 0x88:
      regA = add8(regA, regB, c);
      break;
    case 0x89:
      regA = add8(regA, regC, c);
      break;
    case 0x8A:
      regA = add8(regA, regD, c);
      break;
    case 0x8B:
      regA = add8(regA, regE, c);
      break;
    case 0x8C:
      regA = add8(regA, regH, c);
      break;
    case 0x8D:
      regA = add8(regA, regL, c);
      break;
    case 0x8E:
      regA = add8(regA, ram[regHL], c);
      break;
    case 0x8F:
      regA = add8(regA, regA, c);
      break;
    default:
      break;
  }
}

void do_sub(unsigned char opcode) {
  unsigned char c;

  c = regF & flagC;
  switch (opcode) {
    case 0x90:
      regA = sub8(regA, regB, 0);
      break;
    case 0x91:
      regA = sub8(regA, regC, 0);
      break;
    case 0x92:
      regA = sub8(regA, regD, 0);
      break;
    case 0x93:
      regA = sub8(regA, regE, 0);
      break;
    case 0x94:
      regA = sub8(regA, regH, 0);
      break;
    case 0x95:
      regA = sub8(regA, regL, 0);
      break;
    case 0x96:
      regA = sub8(regA, ram[regHL], 0);
      break;
    case 0x97:
      regA = sub8(regA, regA, 0);
      break;
    case 0x98:
      regA = sub8(regA, regB, c);
      break;
    case 0x99:
      regA = sub8(regA, regC, c);
      break;
    case 0x9A:
      regA = sub8(regA, regD, c);
      break;
    case 0x9B:
      regA = sub8(regA, regE, c);
      break;
    case 0x9C:
      regA = sub8(regA, regH, c);
      break;
    case 0x9D:
      regA = sub8(regA, regL, c);
      break;
    case 0x9E:
      regA = sub8(regA, ram[regHL], c);
      break;
    case 0x9F:
      regA = sub8(regA, regA, c);
      break;
    default:
      break;
  }
}

void do_ana(unsigned char opcode) {
  switch (opcode) {
    case 0xA0:
      regA = and8(regA, regB);
      break;
    case 0xA1:
      regA = and8(regA, regC);
      break;
    case 0xA2:
      regA = and8(regA, regD);
      break;
    case 0xA3:
      regA = and8(regA, regE);
      break;
    case 0xA4:
      regA = and8(regA, regH);
      break;
    case 0xA5:
      regA = and8(regA, regL);
      break;
    case 0xA6:
      regA = and8(regA, ram[regHL]);
      break;
    case 0xA7:
      regA = and8(regA, regA);
      break;
    default:
      break;
  }
}

void do_xra(unsigned char opcode) {
  switch (opcode) {
    case 0xA8:
      regA = xor8(regA, regB);
      break;
    case 0xA9:
      regA = xor8(regA, regC);
      break;
    case 0xAA:
      regA = xor8(regA, regD);
      break;
    case 0xAB:
      regA = xor8(regA, regE);
      break;
    case 0xAC:
      regA = xor8(regA, regH);
      break;
    case 0xAD:
      regA = xor8(regA, regL);
      break;
    case 0xAE:
      regA = xor8(regA, ram[regHL]);
      break;
    case 0xAF:
      regA = xor8(regA, regA);
      break;
    default:
      break;
  }
}

void do_ora(unsigned char opcode) {
  switch (opcode) {
    case 0xB0:
      regA = or8(regA, regB);
      break;
    case 0xB1:
      regA = or8(regA, regC);
      break;
    case 0xB2:
      regA = or8(regA, regD);
      break;
    case 0xB3:
      regA = or8(regA, regE);
      break;
    case 0xB4:
      regA = or8(regA, regH);
      break;
    case 0xB5:
      regA = or8(regA, regL);
      break;
    case 0xB6:
      regA = or8(regA, ram[regHL]);
      break;
    case 0xB7:
      regA = or8(regA, regA);
      break;
    default:
      break;
  }
}

void do_cmp(unsigned char opcode) {
  switch (opcode) {
    case 0xB8:
      sub8(regA, regB, 0);
      break;
    case 0xB9:
      sub8(regA, regC, 0);
      break;
    case 0xBA:
      sub8(regA, regD, 0);
      break;
    case 0xBB:
      sub8(regA, regE, 0);
      break;
    case 0xBC:
      sub8(regA, regH, 0);
      break;
    case 0xBD:
      sub8(regA, regL, 0);
      break;
    case 0xBE:
      sub8(regA, ram[regHL], 0);
      break;
    case 0xBF:
      sub8(regA, regA, 0);
      break;
    default:
      break;
  }
}

void do_ret(unsigned char opcode) {
  unsigned char high, low;
  switch (opcode) {
    case 0xC0:
      if (regF & flagZ) return;
      break;
    case 0xC8:
      if (!(regF & flagZ)) return;
      break;
    case 0xD0:
      if (regF & flagC) return;
      break;
    case 0xD8:
      if (!(regF & flagC)) return;
      break;
    case 0xE0:
      if (regF & flagP) return;
      break;
    case 0xE8:
      if (!(regF & flagP)) return;
      break;
    case 0xF0:
      if (regF & flagS) return;
      break;
    case 0xF8:
      if (!(regF & flagS)) return;
      break;
    default:
      return;
  }
  low = ram[regSP++];
  high = ram[regSP++];
  regPC = (high << 8) + low;
}

void do_pop(unsigned char opcode) {
  unsigned char high, low;
  switch (opcode) {
    case 0xC1:
      regC = ram[regSP++];
      regB = ram[regSP++];
      break;
    case 0xD1:
      regE = ram[regSP++];
      regD = ram[regSP++];
      break;
    case 0xE1:
      regL = ram[regSP++];
      regH = ram[regSP++];
      break;
    case 0xF1:
      regF = ram[regSP++];
      regA = ram[regSP++];
      regF |= 2;
      regF &= 0xD7;
      break;
    case 0xC9:
    case 0xD9:
      low = ram[regSP++];
      high = ram[regSP++];
      regPC = (high << 8) + low;
      break;
    case 0xE9:
      regPC = regHL;
      break;
    case 0xF9:
      regSP = regHL;
    default:
      break;
  }
}

void do_jp(unsigned char opcode) {
  unsigned char high, low;

  low = fetch();
  high = fetch();
  switch (opcode) {
    case 0xC2:
      if (regF & flagZ) return;
      break;
    case 0xCA:
      if (!(regF & flagZ)) return;
      break;
    case 0xD2:
      if (regF & flagC) return;
      break;
    case 0xDA:
      if (!(regF & flagC)) return;
      break;
    case 0xE2:
      if (regF & flagP) return;
      break;
    case 0xEA:
      if (!(regF & flagP)) return;
      break;
    case 0xF2:
      if (regF & flagS) return;
      break;
    case 0xFA:
      if (!(regF & flagS)) return;
      break;
    default:
      return;
  }
  regPC = (high << 8) + low;
}

void do_jmp(unsigned char opcode) {
  unsigned char high, low;
  unsigned short addr;
  switch (opcode) {
    case 0xC3:
    case 0xCB:
      low = fetch();
      high = fetch();
      addr = (high << 8) + low;
      regPC = addr;
      break;
    case 0xD3:
      low = fetch();
      io[low] = regA;
      break;
    case 0xDB:
      low = fetch();
      regA = io[low];
      break;
    case 0xE3:
      low = ram[regSP];
      high = ram[regSP + 1];
      ram[regSP] = regL;
      ram[regSP + 1] = regH;
      regL = low;
      regH = high;
      break;
    case 0xEB:
      low = regE;
      high = regD;
      regE = regL;
      regD = regH;
      regL = low;
      regH = high;
      break;
    case 0xF3:
      intr_flag = 0;
      break;
    case 0xFB:
      intr_flag = 1;
    default:
      return;
  }
}

void do_cz(unsigned char opcode) {
  unsigned char low, high;

  low = fetch();
  high = fetch();
  switch (opcode) {
    case 0xC4:
      if (regF & flagZ) return;
      break;
    case 0xCC:
      if (!(regF & flagZ)) return;
      break;
    case 0xD4:
      if (regF & flagC) return;
      break;
    case 0xDC:
      if (!(regF & flagC)) return;
      break;
    case 0xE4:
      if (regF & flagP) return;
      break;
    case 0xEC:
      if (!(regF & flagP)) return;
      break;
    case 0xF4:
      if (regF & flagS) return;
      break;
    case 0xFC:
      if (!(regF & flagS)) return;
      break;
    default:
      return;
  }
  ram[--regSP] = (unsigned char)(regPC >> 8);
  ram[--regSP] = (unsigned char)(regPC & 0xFF);
  regPC = (high << 8) + low;
}

void do_push(unsigned char opcode) {
  unsigned char low, high;
  switch (opcode) {
    case 0xC5:
      ram[--regSP] = regB;
      ram[--regSP] = regC;
      break;
    case 0xD5:
      ram[--regSP] = regD;
      ram[--regSP] = regE;
      break;
    case 0xE5:
      ram[--regSP] = regH;
      ram[--regSP] = regL;
      break;
    case 0xF5:
      ram[--regSP] = regA;
      ram[--regSP] = regF;
      break;
    case 0xCD:
    case 0xDD:
    case 0xED:
    case 0xFD:
      low = fetch();
      high = fetch();
      ram[--regSP] = (unsigned char)(regPC >> 8);
      ram[--regSP] = (unsigned char)(regPC & 0xFF);
      regPC = (high << 8) + low;
      break;
    default:
      break;
  }
}

void do_adi(unsigned char opcode) {
  unsigned char data;
  unsigned char c = 0;

  data = fetch();
  c = regF & flagC;
  switch (opcode) {
    case 0xC6:
      regA = add8(regA, data, 0);
      break;
    case 0xCE:
      regA = add8(regA, data, c);
      break;
    case 0xD6:
      regA = sub8(regA, data, 0);
      break;
    case 0xDE:
      regA = sub8(regA, data, c);
      break;
    case 0xE6:
      regA = and8(regA, data);
      break;
    case 0xEE:
      regA = xor8(regA, data);
      break;
    case 0xF6:
      regA = or8(regA, data);
      break;
    case 0xFE:
      sub8(regA, data, 0);
      break;
    default:
      break;
  }
}

void do_rst(unsigned char opcode) {
  unsigned char low, high;
  ram[--regSP] = (unsigned char)(regPC >> 8);
  ram[--regSP] = (unsigned char)(regPC & 0xFF);
  regPC = 0x0000 + (opcode - 0xC7);
}

void exec(unsigned char opcode) {
  if ((opcode >= 0) && (opcode < 0x40)) {
    switch (opcode & 0xF) {
      case 0x0:
      case 0x8:
	return do_nop(opcode);
      case 0x1:
	return do_lxi(opcode);
      case 0x2:
	return do_sta(opcode);
      case 0x3:
	return do_inx(opcode);
      case 0x4:
      case 0xC:
	return do_inr(opcode);
      case 0x5:
      case 0xD:
	return do_dcr(opcode);
      case 0x6:
      case 0xE:
	return do_mvi(opcode);
      case 0x7:
      case 0xF:
	return do_rlc(opcode);
      case 0x9:
	return do_dad(opcode);
      case 0xA:
	return do_lda(opcode);
      case 0xB:
	return do_dcx(opcode);
      default:
	return;
    }
  } else if ((opcode >= 0x40) && (opcode < 0x80)) {
    if (opcode < 0x48) return do_movb(opcode);
    if (opcode < 0x50) return do_movc(opcode);
    if (opcode < 0x58) return do_movd(opcode);
    if (opcode < 0x60) return do_move(opcode);
    if (opcode < 0x68) return do_movh(opcode);
    if (opcode < 0x70) return do_movl(opcode);
    if (opcode < 0x78) return do_movm(opcode);
    if (opcode < 0x80) return do_mova(opcode);
  } else if ((opcode >= 0x80) && (opcode < 0xC0)) {
    if (opcode < 0x90) return do_add(opcode);
    if (opcode < 0xA0) return do_sub(opcode);
    if (opcode < 0xA8) return do_ana(opcode);
    if (opcode < 0xB0) return do_xra(opcode);
    if (opcode < 0xB8) return do_ora(opcode);
    if (opcode < 0xC0) return do_cmp(opcode);
  } else {
    switch (opcode & 0xF) {
      case 0x0:
      case 0x8:
	return do_ret(opcode);
      case 0x1:
      case 0x9:
	return do_pop(opcode);
      case 0x2:
      case 0xA:
	return do_jp(opcode);
      case 0x3:
      case 0xB:
	return do_jmp(opcode);
      case 0x4:
      case 0xC:
	return do_cz(opcode);
      case 0x5:
      case 0xD:
	return do_push(opcode);
      case 0x6:
      case 0xE:
	return do_adi(opcode);
      case 0x7:
      case 0xF:
	return do_rst(opcode);
      default:
	return;
    }
  }
}
